import copy
import time
import errno
import socket
import threading

from hks_pylib.logger import LoggerGenerator
from hks_pylib.cryptography.ciphers.cipherid import CipherID
from hks_pylib.cryptography.ciphers.hkscipher import HKSCipher
from hks_pylib.logger.logger_generator import InvisibleLoggerGenerator
from hks_pylib.logger.standard import StdLevels, StdUsers
from hkserror.hkserror import HFormatError, HTypeError

from hks_pynetwork.packet_buffer import PacketBuffer
from hks_pynetwork.secure_packet import SecurePacketEncoder, SecurePacketDecoder

from hks_pylib.errors.cryptography.ciphers import CipherParameterError
from hks_pylib.errors.cryptography.ciphers.symmetrics import UnAuthenticatedPacketError

from hks_pynetwork.errors.packet import IncompletePacketError
from hks_pynetwork.errors.secure_packet import CipherTypeMismatchError
from hks_pynetwork.errors.external import STCPSocketClosedError, STCPSocketTimeoutError


class STCPSocket(object):
    DEFAULT_TIME_OUT = 0.1
    DEFAULT_RELOAD_TIME = 0.1

    def __init__(
                    self,
                    cipher: HKSCipher,
                    name: str,
                    buffer_size: int,
                    logger_generator: LoggerGenerator = InvisibleLoggerGenerator(),
                    display: dict = {}
                ):
        if not isinstance(cipher, HKSCipher):
            raise HTypeError("cipher", cipher, HKSCipher)

        if not isinstance(buffer_size, int):
            raise HTypeError("buffer_size", int)

        if buffer_size <= 0:
            raise HFormatError("Parameter buffer_size expected an positive integer.")
        
        if not isinstance(logger_generator, LoggerGenerator):
            raise HTypeError("logger_generator", logger_generator, LoggerGenerator)

        if not isinstance(display, dict):
            raise HTypeError("display", display, dict)

        if not isinstance(name, str):
            raise HTypeError("name", name, str)

        self._name = name
        self._logger_generator = logger_generator
        self._display = display
        self._log = self._logger_generator.generate(name, self._display)

        self._log(StdUsers.DEV, StdLevels.DEBUG, "Initialized with "
        "cipher {}.".format(CipherID.cls2name(type(cipher))))

        self._socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.__cipher = cipher
        self.__cipher.reset()

        self.set_reload_time(STCPSocket.DEFAULT_RELOAD_TIME)
        self.set_recv_timeout(None)

        self.__packet_encoder = SecurePacketEncoder(self.__cipher)
        self.__packet_decoder = SecurePacketDecoder(self.__cipher)
        self.__buffer = None
        self.__buffer_size = buffer_size

        self.__buffer_available = threading.Event()
        self._stop_auto_recv = False
        self._prepare_close = False

        self._is_working = False

    def __enter__(self):
        return self

    def __exit__(self, *args):
        if self._is_working:
            self.close()

    def _start_auto_recv(self):
        self._log(StdUsers.DEV, StdLevels.INFO, "Start the automatic received process.")
        while True:
            try:
                data = self._socket.recv(self.__buffer_size)
            except socket.error as e:
                if isinstance(e, socket.timeout) and not self._stop_auto_recv:
                    continue
 
                self._socket.close()
                if e.errno in (errno.ECONNRESET, errno.ECONNABORTED, errno.ECONNREFUSED):
                    self._log(StdUsers.DEV, StdLevels.INFO, "Automatic received "
                    "process closed normally ({}).".format(e))
                    break
                elif isinstance(e, socket.timeout):
                    self._log(StdUsers.DEV, StdLevels.INFO, "Automatic received "
                    "process closed normally (timeout and socket closed).")
                    break
                elif isinstance(e, OSError) and e.errno == 10038:
                    self._log(StdUsers.DEV, StdLevels.INFO, "Automatic received "
                    "process closed normally (remote socket closed).")
                    break
                else:
                    self._log(StdUsers.DEV, StdLevels.ERROR, "Automatic received "
                    "process closed with an unknown socket error ({}).".format(e))
                    raise e
            except Exception as e:
                self._socket.close()
                self._log(StdUsers.DEV, StdLevels.ERROR, "Automatic received process "
                "closed with an unknown error ({}).".format(e))
                break
            else:  # If there is no error
                # When be closed by the remote party,
                # socket will receive infinite empty packets
                if not data:
                    self._socket.close()
                    self._log(StdUsers.DEV, StdLevels.INFO, "Automatic received process "
                    "closed normally (remote socket closed).")
                    break
                self.__buffer.push(data)
                self.__buffer_available.set()

        self.__buffer_available.set()

    def set_reload_time(self, value: float):
        if not isinstance(value, (int, float)):
            raise HTypeError("value", value, float, int)

        if value < 0:
            raise HFormatError("Parameter value expected a non-negative number.")

        self._log(StdUsers.DEV, StdLevels.DEBUG, "Set reload "
        "time to {}.".format(value))
        self.__reload_time = value

    def set_recv_timeout(self, value: float):
        if value is not None and not isinstance(value, (int, float)):
            raise HTypeError("value", value, float, int, None)

        if value is not None and value < 0:
            raise HFormatError("Parameter value expected a non-negative number.")

        self._log(StdUsers.DEV, StdLevels.DEBUG, "Set received "
        "timeout to {}.".format(value))
        self.__recv_timeout = value

    def settimeout_raw(self, value: float):
        if value is not None and not isinstance(value, (int, float)):
            raise HTypeError("value", value, float, int, None)

        if value < 0:
            raise HFormatError("Parameter value expected a non-negative number.")

        self._log(StdUsers.DEV, StdLevels.DEBUG, "Set timeout "
        "to {}.".format(value))

        return self._socket.settimeout(value)

    def recv(self) -> bytes:
        data = b''
        if self.__recv_timeout is not None:
            start_recv_time = time.time()

        while not data:
            if self.isclosed() and len(self.__buffer) == 0:
                raise STCPSocketClosedError("Connection closed.")

            try:
                data = self.__buffer.pop()
            except Exception as e:
                self._log(StdUsers.USER, StdLevels.INFO, "Unknown error.")
                self._log(StdUsers.DEV, StdLevels.ERROR, "Unknown error "
                "({}).".format(e))
                break
            finally:
                if not data:
                    if self.__recv_timeout is not None and\
                        time.time() - start_recv_time > self.__recv_timeout:
                        if self.__recv_timeout != 0:
                            raise STCPSocketTimeoutError("Receiving exceeds timeout.")
                        else:
                            raise STCPSocketTimeoutError("Method recv() can not return "
                            "the value immediately.")

                    time.sleep(self.__reload_time)

                    if self.__recv_timeout is None and len(self.__buffer) == 0\
                        and not self.isclosed():
                        self.__buffer_available.wait()

        self.__buffer_available.clear()
        return data

    def send(self, data: bytes) -> int:
        if not isinstance(data, bytes):
            raise HTypeError("data", data, bytes)

        self.__cipher.reset()
        packet = self.__packet_encoder.encode(data)
        return self._socket.send(packet)

    def sendall(self, data: bytes) -> int:
        if not isinstance(data, bytes):
            raise HTypeError("data", data, bytes)

        self.__cipher.reset()
        packet = self.__packet_encoder.encode(data)
        return self._socket.sendall(packet)

    def bind(self, address):
        return self._socket.bind(address)

    def listen(self, __backlog: int = 0):
        self._socket.listen(__backlog)
        self._is_working = True

        self._log(StdUsers.USER, StdLevels.INFO, "Server start listening.")
        self._log(StdUsers.DEV, StdLevels.INFO, "Server start listening.")

    def accept(self):
        socket, addr = self._socket.accept()
        socket.settimeout(STCPSocket.DEFAULT_TIME_OUT)  # timeout for non-blocking socket

        self._log(StdUsers.USER, StdLevels.INFO, "Server accepted {}.".format(addr))
        self._log(StdUsers.DEV, StdLevels.INFO, "Server accepted {}.".format(addr))

        socket = self._fromsocket(socket, addr, start_serve=True)
        socket._is_working = True
        
        return socket, addr

    def connect(self, address):
        self._socket.connect(address)
        self._is_working = True

        self._log(StdUsers.USER, StdLevels.INFO, "Connect to "
        "server {} successfully.".format(address))
    
        self._log(StdUsers.DEV, StdLevels.INFO, "Connect to "
        "server {} successfully.".format(address))
    
        self._log(StdUsers.DEV, StdLevels.DEBUG, "Transform "
        "to STCP Socket {}.".format(address))

        self._log = self._logger_generator.generate(
                f"STCP Socket {address}", self._display)

        self.settimeout_raw(STCPSocket.DEFAULT_TIME_OUT)

        self._stop_auto_recv = False
        auto_recv = threading.Thread(
                target=self._start_auto_recv,
                name="AutoRecv of {}".format(self._name)
            )
        auto_recv.start()

        self.__buffer = PacketBuffer(
                decoder=self.__packet_decoder,
                name="Packet Buffer of {}".format(address),
                logger_generator=self._logger_generator,
                display=self._display
            )

    def close(self):
        if self.__buffer is not None:  # not STCP Listener
            self.__buffer_available.set()
            self._stop_auto_recv = True

        if not self.isclosed():
            self._socket.close()
            self._is_working = False
            self._log(StdUsers.DEV, StdLevels.INFO, "Closed.")

    def isclosed(self):
        "This method returns the raw socket._closed"
        return self._socket._closed

    def isworking(self):
        return self._is_working and not self.isclosed()

    def _fromsocket(self, socket: socket.socket, address, start_serve=True):
        cipher = copy.copy(self.__cipher)

        new_socket = STCPSocket(
                cipher=cipher, 
                buffer_size=self.__buffer_size,
                logger_generator=self._logger_generator,
                display=self._display,
                name=f"STCP Socket {address}"
            )

        new_socket._socket = socket

        new_socket.__buffer = PacketBuffer(
                decoder=new_socket.__packet_decoder,
                name="PacketBuffer of {}".format(address),
                logger_generator=self._logger_generator,
                display=self._display
            )

        if start_serve:
            new_socket._stop_auto_recv = False
            auto_recv = threading.Thread(
                    target=new_socket._start_auto_recv,
                    name="AutoRecv of {}".format(new_socket._name)
                )
            auto_recv.start()

        return new_socket



from hks_pynetwork.errors import HKSPyNetworkError


class InternalError(HKSPyNetworkError):
    "The exception is raised by failures in internal module."


class ChannelError(InternalError):
    "The exception is raised by failures in channel."


class ChannelSlotError(ChannelError):
    "The exception is raised when an error occurs due to slot."


class ChannelClosedError(ChannelError):
    "The exception is raised when the method is working after channel closing."

class ForwardNodeError(ChannelError):
    "The exception is raised by failures in forwarder."



from hks_pynetwork.errors import HKSPyNetworkError


class PacketError(HKSPyNetworkError):
    "The exception is raised by failures in packet module."


class PacketEncodingError(PacketError):
    "The exception raised by some errors in packet encoding."


class PacketDecodingError(PacketError):
    "The exception is raised by some errors in packet decoding."


class PacketSizeError(PacketDecodingError, PacketEncodingError):
    "The exception raised by the invalid size of packet elements"


class IncompletePacketError(PacketDecodingError):
    "The exception is raised when the incomplete packet is extracted."



from os import name
import random
import threading

from hks_pylib.logger import LoggerGenerator
from hks_pylib.logger.logger_generator import InvisibleLoggerGenerator
from hks_pylib.logger.standard import StdLevels, StdUsers
from hkserror.hkserror import HTypeError
from hks_pynetwork.external import STCPSocket, STCPSocketClosedError

from hks_pynetwork.errors.internal import ChannelError, ChannelSlotError, ChannelClosedError, ForwardNodeError


class ChannelBuffer(object):
    def __init__(self):
        self._buffer = []
        self.__lock = threading.Lock()

    def push(self, source: str, message: bytes, obj: object = None):
        self.__lock.acquire()
        self._buffer.append({"source": source, "message": message, "obj": obj})
        self.__lock.release()

    def pop(self, source: str = None):
        self.__lock.acquire()
        if len(self._buffer) == 0:
            return None, None, None

        idx = 0
        if source is not None:
            for i, packet in enumerate(self._buffer):
                if packet["source"] == source:
                    idx = i
                    break

        packet = self._buffer[idx]
        del self._buffer[idx]

        self.__lock.release()
        return packet["source"], packet["message"], packet["obj"]

    def __len__(self):
        return len(self._buffer)


class LocalNode(object):
    nodes = []
    node_names = []
    MAX_NODES = 8
    lock = threading.Lock()

    def __init__(self,
            name: str = None,
            logger_generator: LoggerGenerator = InvisibleLoggerGenerator(),
            display: dict = {}
        ):
        if name is not None and not isinstance(name, str):
            raise HTypeError("name", name, str, None)

        if not isinstance(logger_generator, LoggerGenerator):
            raise HTypeError("logger_generator", logger_generator, LoggerGenerator)

        if not isinstance(display, dict):
            raise HTypeError("display", display, dict)

        if name is None:
            while name is None or name in LocalNode.node_names:
                name = str(random.randint(1000000, 9999999))

        if name in LocalNode.node_names:
            raise ChannelSlotError(f"Name {name} is in use.")

        if len(LocalNode.nodes) >= LocalNode.MAX_NODES:
            raise ChannelSlotError("No available slot in Local Node list.")

        LocalNode.lock.acquire()
        LocalNode.node_names.append(name)
        LocalNode.nodes.append(self)
        LocalNode.lock.release()

        self.name = name
        self._buffer = ChannelBuffer()
        self._closed = False
 
        self._buffer_available = threading.Event()
        self.__send_lock = threading.Lock()
        self.__recv_lock = threading.Lock()

        self._log = logger_generator.generate(
                name,
                display
            )

        self._log(StdUsers.DEV, StdLevels.INFO,
        "{} join to Local Nodes.".format(name))

    def send(self, destination_name: str, message: bytes, obj: object = None):
        if not isinstance(destination_name, str):
            raise HTypeError("destination_name", destination_name, str)

        if not isinstance(message, bytes):
            raise HTypeError("message", message, bytes)

        self.__send_lock.acquire()
        if isinstance(message, bytes) is False:
            raise Exception("Message must be a bytes object.")

        if self._closed:
            raise ChannelClosedError("Channel closed.")

        try:
            slot_of_destination = LocalNode.node_names.index(destination_name)
            destination_node: LocalNode = LocalNode.nodes[slot_of_destination]
        except ValueError:
            raise ChannelSlotError(f"Channel name {destination_name} doesn't exist.")

        destination_node._buffer.push(self.name, message, obj)
        destination_node._buffer_available.set()
        self.__send_lock.release()

    def recv(self, source: str = None):
        if source is not None and not isinstance(source, str):
            raise HTypeError("source", source, str, None)

        if source is not None and source not in LocalNode.node_names:
            raise ChannelError("{} has not existed in Local Nodes.".format(source))

        if self._closed:
            raise ChannelClosedError("Channel closed.")

        if len(self._buffer) == 0 and not self._closed:
            self._buffer_available.wait()
        self._buffer_available.clear()

        if self._closed:
            raise ChannelClosedError("Channel closed.")

        source, msg, obj = self._buffer.pop(source)
        return source, msg, obj

    def close(self):
        if not self._closed:
            self.__recv_lock.acquire()

            self._closed = True
            self._buffer_available.set()

            my_slot = LocalNode.node_names.index(self.name)
            del LocalNode.node_names[my_slot]
            del LocalNode.nodes[my_slot]
            del self._buffer

            name = self.name
            self.name = None
            self.__recv_lock.release()

            self._log(StdUsers.DEV, StdLevels.INFO,
            "{} leaves Local Nodes.".format(name))


class ForwardNode(LocalNode):
    def __init__(
                    self,
                    node: LocalNode,
                    socket: STCPSocket,
                    name: str,
                    implicated_die: bool = False,
                    logger_generator: LoggerGenerator = InvisibleLoggerGenerator(),
                    display: tuple = {}
                ):
        if node is not None and not isinstance(node, LocalNode):
            raise HTypeError("node", node, LocalNode, None)

        if socket is not None and not isinstance(socket, STCPSocket):
            raise HTypeError("socket", socket, STCPSocket, None)

        if name is not None and not isinstance(name, str):
            raise HTypeError("name", name, str, None)

        if not isinstance(implicated_die, bool):
            raise HTypeError("implicated_die", implicated_die, bool)

        if not isinstance(logger_generator, LoggerGenerator):
            raise HTypeError("logger_generator", logger_generator, LoggerGenerator)

        if not isinstance(display, dict):
            raise HTypeError("display", display, dict)

        self._node = node
        self._socket = socket
        super().__init__(
                name=name,
                logger_generator=logger_generator,
                display=display
            )

        self._implicated_die = implicated_die
        if display is None:
            display = socket._log.display

        self._one_thread_stop = threading.Event()

    def set_node(self, node: LocalNode):
        if not isinstance(node, LocalNode):
            raise HTypeError("node", node, LocalNode)

        self._node = node

    def set_socket(self, socket: STCPSocket):
        if not isinstance(socket, STCPSocket):
            raise HTypeError("socket", socket, STCPSocket)

        self._socket = socket

    def start(self):
        self._log(StdUsers.DEV, StdLevels.INFO, "Start forwarding between local "
        "node and remote node.")
        if not self._node or not self._socket:
            raise ForwardNodeError("The node or the socket has been not provided yet.")

        t1 = threading.Thread(
                target=self._wait_message_from_node,
                name="WaitFromLocal of {}".format(self.name)
            )
        t1.setDaemon(True)
        t1.start()

        t2 = threading.Thread(
                target=self._wait_message_from_remote,
                name="WaitFromRemote of {}".format(self.name)
            )
        t2.setDaemon(True)
        t2.start()

        self._one_thread_stop.wait()
        self.close()

        if self._implicated_die:
            self._node.close()
            self._socket.close()
        else:
            if self._node._buffer_available.is_set() is False:
                self._node._buffer_available.set()

        self._log(StdUsers.DEV, StdLevels.INFO, "Stop completely.")

    def _wait_message_from_remote(self):
        while not self._closed:
            try:
                data = self._socket.recv()
            except STCPSocketClosedError:
                self._log(StdUsers.DEV, StdLevels.INFO, "Forwarding message from "
                "remote node closed normally (remote node closed).")
                break
            except Exception as e:
                self._log(StdUsers.USER, StdLevels.INFO, "Forwarding message from "
                "remote node closed with an unknown error in data receiving.")

                self._log(StdUsers.DEV, StdLevels.ERROR, "Forwarding message from "
                "remote node closed with an unknown error in data receiving ({}).".format(e))
                break

            if data:
                try:
                    super().send(self._node.name, data)
                except ChannelClosedError:
                    self._log(StdUsers.DEV, StdLevels.INFO, "Forwarding message from "
                    "remote node closed normally (local node closed).")
                    break
                except Exception as e:
                    self._log(StdUsers.USER, StdLevels.INFO, "Forwarding message from "
                    "remote node closed with an unknown error in data sending.")

                    self._log(StdUsers.DEV, StdLevels.ERROR, "Forwarding message from "
                    "remote node closed with an unknown error "
                    "in data sending ({}).".format(e))
                    break
        self._one_thread_stop.set()

    def _wait_message_from_node(self):
        while not self._closed:
            try:
                _, message, _ = super().recv()
            except AttributeError as e:  
                # (Old version) After closing forwarder,
                # it doesn't have buffer attribute.
                # It should be remove this exception in the future.
                self._log(StdUsers.DEV, StdLevels.WARNING,
                "AttributeError occurs when closing forward.")
                
                self._log(StdUsers.USER, StdLevels.INFO, "Forwarding message "
                "from local node closed normally (forwarder self-closed).")
                
                self._log(StdUsers.DEV, StdLevels.INFO, "Forwarding message "
                "from local node closed normally (forwarder self-closed).")
                break
            except ChannelClosedError:
                self._log(StdUsers.USER, StdLevels.INFO, "Forwarding message "
                "from local node closed normally (forwarder self-closed).")

                self._log(StdUsers.DEV, StdLevels.INFO, "Forwarding message "
                "from local node closed normally (forwarder self-closed).")
                break
            except Exception as e:
                self._log(StdUsers.USER, StdLevels.INFO, "Forwarding message "
                "from local node closed with an unknown error in data receiving.")

                self._log(StdUsers.DEV, StdLevels.ERROR, "Forwarding message "
                "from local node closed with an unknown "
                "error ({}) in data receiving.".format(e))
                break

            if message:
                try:
                    self._socket.send(message)
                except STCPSocketClosedError:
                    self._log(StdUsers.DEV, StdLevels.INFO, "Forwarding message "
                    "from local node closed normally (remote node closed).")
                    break
                except Exception as e:
                    self._log(StdUsers.USER, StdLevels.INFO, "Forwarding message "
                    "from local node closed with an unknown error in data sending.")

                    self._log(StdUsers.DEV, StdLevels.ERROR, "Forwarding message "
                    "from local node closed with an unknown "
                    "error in data sending ({}).".format(e))
                    break

        self._one_thread_stop.set()

    def send(self, received_node_name, message):
        raise NotImplementedError("Forwarder doesn't have send() method.")

    def recv(self):
        raise NotImplementedError("Forwarder doesn't have recv() method.")



import threading

from hks_pylib.logger import LoggerGenerator
from hks_pylib.logger.logger_generator import InvisibleLoggerGenerator
from hks_pylib.logger.standard import StdLevels, StdUsers
from hkserror.hkserror import HTypeError
from hks_pynetwork.secure_packet import PacketDecoder

from hks_pynetwork.errors.packet import IncompletePacketError, PacketSizeError
from hks_pynetwork.errors.secure_packet import CipherTypeMismatchError


class PacketBuffer():
    def __init__(
                    self,
                    decoder: PacketDecoder,
                    name: str,
                    logger_generator: LoggerGenerator = InvisibleLoggerGenerator(),
                    display: dict = {}
                ) -> None:
        if not isinstance(decoder, PacketDecoder):
            raise HTypeError("decoder", decoder, PacketDecoder)

        if name is not None and not isinstance(name, str):
            raise HTypeError("name", name, str, None)

        self._buffer = []

        self._packet_decoder = decoder

        self.__print = logger_generator.generate(name, display)
        
        self._current_packet = b""
        self._current_packet_size = 0
        self._expected_current_packet_size = 0
        
        self._push_lock = threading.Lock()

    def push(self, packet: bytes, append_to_end: bool = True):
        if not isinstance(packet, bytes):
            raise HTypeError("packet", packet, bytes)

        if not isinstance(append_to_end, bool):
            raise HTypeError("append_to_end", append_to_end, bool)

        self._push_lock.acquire()

        if append_to_end:
            self._buffer.append(packet)
        else:
            self._buffer.insert(0, packet)

        self._push_lock.release()

    def pop(self):
        if len(self._buffer) == 0:
            return b""

        self._current_packet_size += len(self._buffer[0])
        self._current_packet += self._buffer[0]
        del self._buffer[0]

        if self._packet_decoder is None:
            ret = self._current_packet
            self._current_packet = b""
            self._current_packet_size = 0
            self._expected_current_packet_size = 0
            return ret

        if self._expected_current_packet_size == 0:
            try:
                packet_dict = self._packet_decoder.get_header(self._current_packet)
                self._expected_current_packet_size =\
                    packet_dict["payload_size"] + packet_dict["header_size"]
            except IncompletePacketError:
                return b""
            except PacketSizeError:
                self.__print(StdUsers.DEV, StdLevels.WARNING, "Detect an "
                "abnormal packet (invalid size).")

                self._current_packet = b""
                self._current_packet_size = 0
                self._expected_current_packet_size = 0
                return b""

        if self._current_packet_size < self._expected_current_packet_size:
            return b""

        try:
            packet_dict = self._packet_decoder.decode(self._current_packet)
        except CipherTypeMismatchError as e:
            self.__print(StdUsers.DEV, StdLevels.WARNING, "Detect an "
            "abnormal packet ({}).".format(e))
            return b""
        finally:
            if  self._current_packet_size > self._expected_current_packet_size:
                apart_of_next_packet = self._current_packet[
                        self._expected_current_packet_size :
                    ]
                self.push(apart_of_next_packet, append_to_end=False)

            self._current_packet = b""
            self._current_packet_size = 0
            self._expected_current_packet_size = 0

        return packet_dict["payload"]

    def __len__(self):
        return len(self._buffer)



from hks_pynetwork.errors import HKSPyNetworkError


class ExternalError(HKSPyNetworkError):
    "The exception is raised by failures in external module."


class STCPSocketError(ExternalError):
    "The exception is raised by failures in STCPSocket."


class STCPSocketClosedError(STCPSocketError):
    "The exception is raised when a STCPSocket was closed."


class STCPSocketTimeoutError(STCPSocketError):
    "The exception is raised when a STCPSocket method stop due to timeout."
